// DropTarget.cpp: implementation of the CDropTarget class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "DropTarget.h"
#include "Group.h"
#include "utils.h"
#include "pidl.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDropTarget::CDropTarget()
{
	m_pDropHelper = NULL;
	CoCreateInstance(CLSID_DragDropHelper, NULL, CLSCTX_INPROC_SERVER, IID_IDropTargetHelper, (LPVOID*)&m_pDropHelper);
	if (!m_pDropHelper)
	{
		TRACE_LOG("IDropTargetHelper not available...");
	}
}

CDropTarget::~CDropTarget()
{
	if (m_pDropHelper)
	{
		m_pDropHelper->Release();
	}
}

UINT CDropTarget::cf_types[CLIPFORMATS_USED];

void CDropTarget::RegisterFormats(void)
{
	//TRACE_LOG("CDropTarget::RegisterFormats");

	cf_types[0] = RegisterClipboardFormat(CFSTR_SHELLIDLIST);
	cf_types[1] = RegisterClipboardFormat(CFSTR_SHELLIDLISTOFFSET);
}


HRESULT STDMETHODCALLTYPE CDropTarget::QueryInterface( 
	REFIID riid,
	void __RPC_FAR *__RPC_FAR *ppvObject)
{
	//TRACE_LOG("IDropTarget::QueryInterface");

	if (riid == IID_IUnknown ||
		riid == IID_IDropTarget)
	{
		*ppvObject = this;
		AddRef();
		return NOERROR;
	}
	return E_NOINTERFACE;
}


ULONG STDMETHODCALLTYPE CDropTarget::AddRef()
{
	//TRACE_LOG("IDropTarget::AddRef");
	return 1;
}


ULONG STDMETHODCALLTYPE CDropTarget::Release()
{
	//TRACE_LOG("IDropTarget::Release");
	return 1;
}


HRESULT STDMETHODCALLTYPE CDropTarget::DragEnter(
	IDataObject *pDataObj,
	DWORD grfKeyState,
	POINTL ptl,
	DWORD *pdwEffect)
{
	//TRACE_LOG("IDropTarget::DragEnter");

	if (m_pDropHelper)
	{
		POINT pt = {ptl.x, ptl.y};
		m_pDropHelper->DragEnter(m_pGroup->m_hwndView, pDataObj, &pt, *pdwEffect);
	}

	FORMATETC etc;
	ZeroMemory(&etc,sizeof(FORMATETC));

	m_pGroup->SetDropHover(&ptl, pdwEffect);
	if (*pdwEffect == DROPEFFECT_NONE)
	{
		// bad drop point =(
		return S_OK;
	}

	// We want all of the content, in a global handle
	etc.tymed = TYMED_HGLOBAL;
	etc.dwAspect = DVASPECT_CONTENT;
	etc.lindex = -1;

	etc.cfFormat = cf_types[0];
	if (pDataObj->QueryGetData(&etc) != S_OK)
	{
		*pdwEffect = DROPEFFECT_NONE;
	}
	return S_OK;
}


HRESULT STDMETHODCALLTYPE CDropTarget::DragOver( 
	DWORD grfKeyState,
	POINTL ptl,
	DWORD *pdwEffect)
{
	//TRACE_LOG("IDropTarget::DragOver");

	if (m_pDropHelper)
	{
		POINT pt = { ptl.x, ptl.y };
		m_pDropHelper->DragOver(&pt, *pdwEffect);
	}

	m_pGroup->SetDropHover(&ptl, pdwEffect);
	return S_OK;
}


HRESULT STDMETHODCALLTYPE CDropTarget::DragLeave()
{
	//TRACE_LOG("IDropTarget::DragLeave");
    if (m_pDropHelper)
	{
		m_pDropHelper->DragLeave();
	}
	m_pGroup->ClearDropHover();
	return S_OK;
}


#define HIDA_GetPIDLParent(pida) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[0])
#define HIDA_GetPIDLItem(pida, i) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[i+1])



HRESULT STDMETHODCALLTYPE CDropTarget::Drop( 
	IDataObject *pDataObj,
	DWORD grfKeyState,
	POINTL ptl,
	DWORD *pdwEffect)
{
	//TRACE_LOG("IDropTarget::Drop");

	FORMATETC etc;
	STGMEDIUM stg1, stg2;
	CIDA *pida;
	POINT *pOffsets;
	bool bExecute;

	if (m_pDropHelper)
	{
		POINT pt = { ptl.x, ptl.y };
		m_pDropHelper->Drop(pDataObj, &pt, *pdwEffect);
	}

	m_pGroup->ClearDropHover();

	ZeroMemory(&etc,sizeof(FORMATETC));
	// We want all of the content, in a global handle
	etc.tymed = TYMED_HGLOBAL;
	etc.dwAspect = DVASPECT_CONTENT;
	etc.lindex = -1;
	etc.cfFormat = cf_types[0];

	IDropTarget *pObject = m_pGroup->GetLastItemTarget(&bExecute);
	if (pObject)
	{
		TRACE_LOG("Object\'s target will be used");
		pObject->DragEnter(pDataObj, grfKeyState, ptl, pdwEffect);
		pObject->DragOver(grfKeyState, ptl, pdwEffect);
		pObject->Drop(pDataObj, grfKeyState, ptl, pdwEffect);
		pObject->Release();
		return S_OK;
	}

	LPSHELLFOLDER pDesktopFolder;
	SHGetDesktopFolder(&pDesktopFolder);
	bool bReorderOnly = false;
	if (pDataObj->GetData(&etc, &stg1) == S_OK)
	{
		pida = reinterpret_cast<CIDA*>(GlobalLock(stg1.hGlobal));
		bReorderOnly = (pDesktopFolder->CompareIDs(0, HIDA_GetPIDLParent(pida), m_pGroup->m_pidlFolder) == 0);
		GlobalUnlock(stg1.hGlobal);
	}
	if (bReorderOnly || bExecute)
	{
		/// Grab the stuff
		if (pDataObj->GetData(&etc, &stg1) == S_OK)
		{
			pida = reinterpret_cast<CIDA*>(GlobalLock(stg1.hGlobal));
			if (pida == NULL)
			{
				TRACE_LOG("Unknown error raised during drag-and-drop processing");
				ReleaseStgMedium(&stg1);
				return E_UNEXPECTED;
			}
			etc.cfFormat = cf_types[1];
			if (pDataObj->GetData(&etc, &stg2) == S_OK)
			{
				pOffsets = reinterpret_cast<POINT*>(GlobalLock(stg2.hGlobal));
			}
			else
			{
				pOffsets = NULL;
			}

			m_pGroup->Drop(grfKeyState, &ptl, pdwEffect, pida, pOffsets);

			// unlock global memory
			GlobalUnlock(stg1.hGlobal);
			if (pOffsets)
			{
				GlobalUnlock(stg2.hGlobal);
			}

			// and free the mediums
			ReleaseStgMedium(&stg1);
			ReleaseStgMedium(&stg2);
		}
	}
	else
	{
		LPITEMIDLIST pidl = (LPITEMIDLIST)CoTaskMemAlloc(sizeof(USHORT));
		ZeroMemory(pidl, sizeof(USHORT));
		m_pGroup->m_pFolder->GetUIObjectOf(NULL, 1, (LPCITEMIDLIST*)&pidl, IID_IDropTarget, 0, (void**)&pObject);
		if (pObject)
		{
			pObject->DragEnter(pDataObj, grfKeyState, ptl, pdwEffect);
			pObject->DragOver(grfKeyState, ptl, pdwEffect);
			pObject->Drop(pDataObj, grfKeyState, ptl, pdwEffect);
			pObject->Release();
		}
		else
		{
			TRACE_LOG("They fucked me");
		}
		CoTaskMemFree(pidl);
		return S_OK;
	}

	return S_OK;
}


HRESULT CDropTarget::Register(CGroup *pGroup)
{
	m_pGroup = pGroup;
	return RegisterDragDrop(m_pGroup->m_hwndListView, this);
}

