// main.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include "main.h"
#include "group.h"
#include "utils.h"
#include "bangs.h"
#include "DropTarget.h"

extern "C" {
    __declspec( dllexport ) int initModuleEx(HWND hwndParent, HINSTANCE hDllInstance, LPCSTR szPath);
    __declspec( dllexport ) void quitModule(HINSTANCE hDllInstance);
}

#define CONFIG_TOKENS_NUM	2

const char g_rcsRevision[]	= "1.0.3";
const char g_szAppName[]	= "Clickonic";
const char g_szMsgHandler[]	= "LSClickonicManager";
const char g_szAuthor[]		= "Seg@, Andymon";

HINSTANCE g_hInstance;

int g_lsMessages[] = {LM_GETREVID, 0};
HWND g_hwndMessageHandler;
void CreateMessageHandler(HINSTANCE hInst);
LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);

// we are adding data to the beginning of the g_pGroups list
// since it is easier to implement
PGROUPLISTITEM g_pGroups = NULL;	


///////////////////////////////////////////////////////////////////////////////
//
// DLL entry point
//
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	if (ul_reason_for_call == DLL_PROCESS_ATTACH)
	{
		DisableThreadLibraryCalls((HINSTANCE)hModule);
	}
	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
//
// Module entry point
//
int initModuleEx(HWND hwndParent, HINSTANCE hDllInstance, LPCSTR szPath)
{
	TRACE_LOG("Initializing");

	CreateMessageHandler(hDllInstance);

	OleInitialize(NULL);
	CoInitialize(NULL); // initialize COM

	InitCommonControls();
	CDropTarget::RegisterFormats();

	FILE *f;
	if (!(f = LCOpen(NULL)))
	{
		utils::ErrorBox("Error: Cannot open step.rc for reading");
	}
	else
	{
		char szLine[MAX_LINE_LENGTH];	// current line
		// let's prepare string array for tokens
		char token[CONFIG_TOKENS_NUM][MAX_LINE_LENGTH];
		char* tokens[CONFIG_TOKENS_NUM];
		char szExtra[MAX_LINE_LENGTH];
		int i;

		for (i = 0; i<CONFIG_TOKENS_NUM; i++)
		{
			tokens[i] = token[i];
		}

		char szConfigLine[MAX_RCCOMMAND];
		StringCchPrintf(szConfigLine, MAX_RCCOMMAND, "*%s", g_szAppName);
		while(LCReadNextConfig(f, szConfigLine, szLine, MAX_LINE_LENGTH))
		{
			LCTokenize(szLine, tokens, CONFIG_TOKENS_NUM, szExtra);
			if (token[1][0] == 0)
			{
				utils::ErrorBoxPrint("Invalid config: *%s string without group name found", g_szAppName);
			}
			else
			{
				CGroup *pGroup = new CGroup(token[1]);
				if (!pGroup->m_bInitialized)
				{
					delete pGroup;
				}
				else
				{
					PGROUPLISTITEM pItem = new GroupListItem;
					pItem->pGroup = pGroup;
					pItem->pNext = g_pGroups;
					g_pGroups = pItem;

					pGroup->activateWndProc();
				}
			}
		}
		LCClose(f);
		RegisterBangs();
	}
	TRACE_LOG("Module has been started");
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//
// Module destructor
//
void quitModule(HINSTANCE hDllInstance)
{
	TRACE_LOG("Destructor called");

	UnregisterBangs();

	while (g_pGroups)
	{
		PGROUPLISTITEM pFirst = g_pGroups;
		g_pGroups = g_pGroups->pNext;
		delete pFirst->pGroup;
		delete pFirst;
	}

	if (g_hwndMessageHandler)
	{
		SendMessage(GetLitestepWnd(),LM_UNREGISTERMESSAGE,(WPARAM)g_hwndMessageHandler, (LPARAM)g_lsMessages);
		DestroyWindow(g_hwndMessageHandler);
	}
	UnregisterClass(g_szMsgHandler, hDllInstance);

	CoUninitialize(); // shut down COM
	OleUninitialize();

	TRACE_LOG("Module has been unloaded");
}


///////////////////////////////////////////////////////////////////
//
// Find the icon group with the specified name
//
CGroup* GetGroupByName(const char *szName)
{
	char szBuf[MAX_LINE_LENGTH];
	StringCchCopy(szBuf, MAX_LINE_LENGTH, szName);
	PathUnquoteSpaces(szBuf);
	PGROUPLISTITEM pEnumerator = g_pGroups;
	while (pEnumerator)
	{
		if (stricmp(pEnumerator->pGroup->m_szName, szBuf) == 0)
		{
			return pEnumerator->pGroup;
		}
		pEnumerator = pEnumerator->pNext;
	}
	return NULL;
}


////////////////////////////////////////////////////////////////////////////
//
// Create message handler window
//

void CreateMessageHandler(HINSTANCE hInst)
{
	WNDCLASSEX wc;
	ZeroMemory(&wc, sizeof(WNDCLASSEX));
	wc.cbSize = sizeof(WNDCLASSEX);
	wc.lpfnWndProc = MessageHandlerProc;
	wc.hInstance = hInst;
	wc.lpszClassName = g_szMsgHandler;
	wc.hIconSm = 0;

	if (!RegisterClassEx(&wc))
	{
		utils::ErrorBox("Error: Cannot register handler window class...");
		return;
	}

	g_hwndMessageHandler = CreateWindowEx(WS_EX_TOOLWINDOW,
		g_szMsgHandler,
		0,
		WS_POPUP,
		0, 0, 0, 0, 
		0,
		0,
		hInst,
		0);

	if (!g_hwndMessageHandler)
	{
		utils::ErrorBox("Error: Cannot create main window...");
		return;
	}	
	SendMessage(GetLitestepWnd(), LM_REGISTERMESSAGE, (WPARAM)g_hwndMessageHandler, (LPARAM) g_lsMessages);

}


////////////////////////////////////////////////////////////////////////////
//
// Message handler process the following messages
//	LM_GETREVID				-	Revision ID
//
LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
	case LM_GETREVID:
		LPSTR buf;
		buf = (LPSTR)lParam;
		switch (wParam)
		{
		case 0:
		case 1:
			StringCchPrintf(buf, lParam, "%s: %s (%s)", g_szAppName, g_rcsRevision, g_szAuthor);
			break;
		default:
			StringCchCopy(buf, lParam, "");
		}
		return strlen(buf);
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}


LRESULT WINAPI ListView_Proc(HWND hListView, UINT msg, WPARAM wParam, LPARAM lParam)
{
	CGroup* group = (CGroup *) GetWindowLong(hListView, GWL_USERDATA);

	if ( group )
	{
		if ( group->m_bExplorerDesktop )
			return CallWindowProc(group->wpOrigListView_Proc, hListView, msg, wParam, lParam);

		switch(msg)
		{
			case WM_LBUTTONDOWN:
			case WM_LBUTTONUP:
			{
				if ( (0x8000 & GetKeyState(VK_CONTROL)) || (0x8000 & GetKeyState(VK_SHIFT)) || (0x8000 & GetKeyState(VK_MENU)) )
				{
					LVHITTESTINFO lvhi;
					lvhi.flags = LVHT_ONITEM;
					lvhi.pt.x = GET_X_LPARAM(lParam);
					lvhi.pt.y = GET_Y_LPARAM(lParam);

					if ( ListView_HitTest(group->m_hwndListView, &lvhi) != -1 )
						return CallWindowProc(group->wpOrigListView_Proc, hListView, msg, wParam, lParam);

					group->relayMouseMessage(msg, wParam, lParam);
					return 0;
				}
			}
			break;

			case WM_RBUTTONDOWN:
			case WM_RBUTTONUP:
			{
				LVHITTESTINFO lvhi;
				lvhi.flags = LVHT_ONITEM;
				lvhi.pt.x = GET_X_LPARAM(lParam);
				lvhi.pt.y = GET_Y_LPARAM(lParam);
						
				if ( ListView_HitTest(group->m_hwndListView, &lvhi) != -1 )
					return CallWindowProc(group->wpOrigListView_Proc, hListView, msg, wParam, lParam);

				group->relayMouseMessage(msg, wParam, lParam);
				return 0;
			}
			break;

			case WM_MBUTTONDOWN:
			case WM_MBUTTONUP:
			{
				LVHITTESTINFO lvhi;
				lvhi.flags = LVHT_ONITEM;
				lvhi.pt.x = GET_X_LPARAM(lParam);
				lvhi.pt.y = GET_Y_LPARAM(lParam);
						
				if ( ListView_HitTest(group->m_hwndListView, &lvhi) != -1 )
					return CallWindowProc(group->wpOrigListView_Proc, hListView, msg, wParam, lParam);

				group->relayMouseMessage(msg, wParam, lParam);
				return 0;
			}
			break;
		}

		return CallWindowProc(group->wpOrigListView_Proc, hListView, msg, wParam, lParam);
	}
	else
	{
		return DefWindowProc(hListView, msg, wParam, lParam);
	}
}